/*
 Copyright (C) 2010-2024 by Hoefer Alfred

 Licensed under the LGPL v2.1, see the file COPYING.LIB
*/

#include "image.h"
#include <setjmp.h>
#include "../image_pv.h"

#define USE_INBUFER 1

/*
Magic Number	Dateityp	Kodierung
P1	Portable Bitmap		ASCII
P2	Portable Graymap	ASCII
P3	Portable Pixmap		ASCII
P4	Portable Bitmap		Bin
P5	Portable Graymap	Bin
P6	Portable Pixmap		Bin

Portable Bitmap  *.pbm	0 und 1				1 Bit
Portable Graymap *.pgm	0 ... 255 bzw. 0 ... 65535	8 Bit bzw. 16 Bit
Portable Pixmap  *.ppm	0 ... 255 bzw. 0 ... 65535	24 bzw. 48 Bit

--- Only *.pam (Netpbm Formats) save ALPHA ----
P7
WIDTH 227
HEIGHT 149
DEPTH 3
MAXVAL 255
TUPLTYPE RGB
ENDHDR
*/

typedef struct TtmpIO {
#if USE_INBUFER
  int idx;
  int incount;
#endif
  int fu;	//1=buffer on
  const char *err_txt;
  jmp_buf setjmp_buffer;
  byte txt[32];
#if USE_INBUFER
  byte line[BUFSIZ];
#endif
} TtmpIO;

static void read_DATA(TImage *self, void *data, int lg) {
	if (fread(data, 1, lg, self->in_fp) != lg) 
			longjmp(self->tIO->setjmp_buffer, 1);
}

#if USE_INBUFER
static dword buf_getc(TImage *self) {
	if (self->tIO->incount == EOF) return EOF;
	if (self->tIO->idx >= self->tIO->incount) {
	    self->tIO->incount = fread(self->tIO->line, 1, BUFSIZ, self->in_fp);
	    if (self->tIO->incount <= 0) {
		self->tIO->incount = EOF;
		return EOF;
	    }
	    self->tIO->idx = 0;
	}
	return self->tIO->line[self->tIO->idx++];
}
#endif

static dword read_VAL(TImage *self) {
  dword value, b;
  bool bskip = false;
  bool valOk = false;
	value = 0;
loop:
#if USE_INBUFER
	if (self->tIO->fu & 1) b = buf_getc(self); else
#endif
	b = getc(self->in_fp);

	if (b == EOF) longjmp(self->tIO->setjmp_buffer, 1);

	if (b <= 0x20) {
	    if ((b == '\n') || (b == '\r')) bskip = false;
	    if (!bskip && valOk) return value;
	    goto loop;
	}
	if (bskip) goto loop;
	if (b == '#') {
	    bskip = true;
	    goto loop;
	}
	if ((b < '0') || (b > '9')) {
	    self->tIO->err_txt = "Format: not unsigned integer";
	    longjmp(self->tIO->setjmp_buffer, 1);
	}
	value = (value * 10) + (b & 0x0f);
	valOk = true;
	goto loop;
}

static dword read_Bool(TImage *self) {
  dword value, b;
  bool bskip = false;
  bool valOk = false;
	value = 0;
loop:
#if USE_INBUFER
	if (self->tIO->fu & 1) b = buf_getc(self); else
#endif
	b = getc(self->in_fp);

	if (b == EOF) longjmp(self->tIO->setjmp_buffer, 1);

	if (b <= 0x20) {
	    if ((b == '\n') || (b == '\r')) bskip = false;
	    if (!bskip && valOk) return value;
	    goto loop;
	}
	if (bskip) goto loop;
	if (b == '#') {
	    bskip = true;
	    goto loop;
	}
	if ((b < '0') || (b > '1')) {
	    self->tIO->err_txt = "Value: not 0 or 1";
	    longjmp(self->tIO->setjmp_buffer, 1);
	}
	value = (b & 0x0f);
	return value;
}

static char *read_TXT(TImage *self) {
  dword b, len;
  byte *pc;
  bool bskip = false;

	pc = self->tIO->txt;
	*pc = 0;
	len = 0;
loop:
	b = getc(self->in_fp);
	if (b == EOF) longjmp(self->tIO->setjmp_buffer, 1);

	if (b <= 0x20) {
	    if ((b == '\n') || (b == '\r')) bskip = false;
	    if (!bskip && (pc > self->tIO->txt)) {
		*pc = 0;
		return (char*)self->tIO->txt;
	    }
	    goto loop;
	}
	if (bskip) goto loop;
	if (b == '#') {
	    bskip = true;
	    goto loop;
	}
	if ((b < ' ') || (b > 'z') || (len >= BUFSIZ)) {
	    self->tIO->err_txt = "Format: is not txt";
	    longjmp(self->tIO->setjmp_buffer, 1);
	}
	*pc = b; pc++;
	len++;
	goto loop;
}

#define PNM_P1 1
#define PNM_P2 2
#define PNM_P3 3
#define PNM_P4 4
#define PNM_P5 5
#define PNM_P6 6
#define PNM_P7 7

int read_pnm(TImage *self) {
  int pnm_maxval, BixByte, pnmType;
  byte *src_buf = NULL;
  TtmpIO *tIO;
  char *pc;
  bool res, is_pbm;

	res = false;
	is_pbm = false;
	tIO = xcalloc(sizeof(TtmpIO));
	self->tIO = tIO;
	tIO->err_txt = "Unknown Format: ";
	if (setjmp(tIO->setjmp_buffer)) {
err:
	    appError(EMSG_ERR1, "%s(%d): %s \"%s\"", FL, tIO->err_txt, self->in_gp->name);
	    goto Close02;
	}

	pnmType = 0;
	BixByte = 0;
#if 1
	pc = read_TXT(self);
	if (!pc || (pc[0] != 'P') ||
	    (pc[1] < '1') || (pc[1] > '7') || 
	    (pc[2] != 0)) goto err;

	switch (pc[1]) {
	case '1':
	    pnmType = PNM_P1;
	    is_pbm = true;
	    break;
	case '2':
	    pnmType = PNM_P2;
	    break;
	case '3':
	    pnmType = PNM_P3;
	    tIO->fu = 1;
	    break;
	case '4':
	    pnmType = PNM_P4;
	    is_pbm = true;
	    break;
	case '5':
	    BixByte = 1;
	    pnmType = PNM_P5;
	    break;
	case '6':
	    BixByte = 3;
	    pnmType = PNM_P6;
	    break;
	case '7':
	    pnmType = PNM_P7;
	    BixByte = 3;
	    break;
	}
#else
  Tic32 ic;

	read_DATA(self, &ic, 2);
	if (ic.w[0] == 0x3150) {
	    pnmType = PNM_P1;
	    is_pbm = true;
	} else if (ic.w[0] == 0x3250) {
	    pnmType = PNM_P2;
//	    is_pgm = true;
	} else if (ic.w[0] == 0x3350) {
	    pnmType = PNM_P3;
	    tIO->fu = 1;
	} else if (ic.w[0] == 0x3450) {
	    pnmType = PNM_P4;
//	    goto err;
	    is_pbm = true;
	} else if (ic.w[0] == 0x3550) {
	    BixByte = 1;
	    pnmType = PNM_P5;
	} else if (ic.w[0] == 0x3650) {
	    BixByte = 3;
	    pnmType = PNM_P6;
	} else if (ic.w[0] == 0x3750) {
	    pnmType = PNM_P7;
	    BixByte = 3;
	} else goto err;
#endif
	pnm_maxval = 0;
	if (pnmType == PNM_P7) {
	    while (1) {
		pc = read_TXT(self);
		if (pccmp(pc, "WIDTH")) {
		    self->width = read_VAL(self);
		} else if (pccmp(pc, "HEIGHT")) {
		    self->height = read_VAL(self);
		} else if (pccmp(pc, "DEPTH")) {
		    BixByte = read_VAL(self);
		} else if (pccmp(pc, "MAXVAL")) {
		    pnm_maxval = read_VAL(self);
		} else if (pccmp(pc, "TUPLTYPE")) {
		    pc = read_TXT(self);
		} else if (pccmp(pc, "ENDHDR")) {
		    goto P7_End;
		} else goto err;
	    }
P7_End:
	    if ((BixByte != 3)&&(BixByte != 4)) goto err;
	    if (BixByte == 3) pnmType = PNM_P6;
	} else {
	    self->width = read_VAL(self);
	    self->height = read_VAL(self);
	    if (!is_pbm) pnm_maxval = read_VAL(self);
	}
// printf("%s(%d): %d %d %d %d\n", FL, self->width, self->height, pnm_maxval, depth);
	if (!is_pbm && !pnm_maxval) goto err;
	if (pnm_maxval == 255) pnm_maxval = 0;
	if (self->pxm_LD & pxmLD_OnlyInfo) goto EndeOk;
	if (!TImage_SetMem(self)) goto err;

//  int lg = self->width * BixByte;
  byte *db, *db0, *sb, *pixAlpha;
  dword A, rdb;
  bool rgb565 = (self->pxm_LD & pxmLD_bppMask) == pxm_2b_16bpp;
  bool ALPHA = (pnmType == PNM_P7) && !(self->pxm_LD & pxmLD_NoMask);

	pixAlpha = NULL;
	if (rgb565 && ALPHA) pixAlpha = xmalloc(self->width * self->height);
	self->rgb565_alpha_data = pixAlpha;
	if (ALPHA) self->stat |= pmxStat_has_alpha;

	A = 255;
	rdb = 0x10000;
	BixByte *= self->width;
	if (BixByte) src_buf = xmalloc(BixByte);
	db0 = self->data;
	for (int y=self->height;y;y--) {
	    if (src_buf) read_DATA(self, src_buf, BixByte);
	    db = db0;
	    sb = src_buf;
	    for (int x=self->width;x;x--) {
  dword R, G, B;
		if (sb) {
		    R = *sb++;
		    if (pnmType == PNM_P5) {
			G = B = R;
		    } else {
			G = *sb++;
			B = *sb++;
			if (pnmType == PNM_P7) {
			    if (ALPHA) A = *sb;
			    sb++;
			}
		    }
		} else {
		    if (pnmType == PNM_P1) {
			R = read_Bool(self);
			if (R) R = 0; else R = 255;
			G = B = R;

		    } else if (pnmType == PNM_P2) {
			R = read_VAL(self);
			G = B = R;

		    } else if (pnmType == PNM_P3) {
			R = read_VAL(self);
			G = read_VAL(self);
			B = read_VAL(self);

		    } else {
			if (rdb & 0x10000) {
			    rdb = getc(self->in_fp);
			    if (rdb == EOF) rdb = 0; else rdb |= 0x0100;
			}
			R = 255;
			if (rdb & 0x80) R = 0;
			G = B = R;
			rdb = rdb << 1;
		    }
		}

		if (pnm_maxval) {
		    R = (R * 255) / pnm_maxval;
		    G = (G * 255) / pnm_maxval;
		    B = (B * 255) / pnm_maxval;
		    A = (A * 255) / pnm_maxval;
		}

		if (rgb565) {
		    *((word*)db) = ((R & 0xf8)<<8)|((G & 0xfc)<<3)|((B & 0xf8)>>3);
		    db += 2;
		    if (pixAlpha) *pixAlpha++ = A;
		} else {
		    *db++ = B;
		    *db++ = G;
		    *db++ = R;
		    *db++ = A;
		}
	    }
	    db0 += self->bytes_per_line;
	}
	alpha_bm_stat(self);
EndeOk:
	res = true;
Close02:
	xfree(src_buf);
	xfree(tIO);
	return res;
}

static bool ChkType(TglobPath *gp, const char *fn, FILE *F) {
  char m[4];

	if (!TImage_ChkPattern(gp, fn, hImageRD_pnm.pattern)) return false;
	if (!F) return true;
	rewind(F);
	if ((fread(&m, 1, 4, F) == 4) &&
	    (m[0] == 'P') && (m[1] >= '1') && (m[1] <= '7') &&
	    ((m[2] == ' ') || (m[2] == '\n') || 
	     (m[2] == '\r') || (m[2] == '#'))) return true;

	return false;
}

const ThImageLoader hImageRD_pnm = {
  "pnm pbm pgm ppm pam",
  filetyp_PNM,
  ChkType,
  read_pnm
};
