/*
 Copyright (C) 2010-2017 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appbase.h"
#include "tthread_pv.h"

static TThread *app_private_thd;
static bool thd_private_EV_lock;

/*_____________________________ TThread _________________________________*/

#ifdef _OS_WIN32
void REGP1 TThread_lock(TThread *self) {
	if (self) EnterCriticalSection(&self->thd_pv->mutex);
}

void REGP1 TThread_unlock(TThread *self) {
	if (self) LeaveCriticalSection(&self->thd_pv->mutex);
}
#else
static pthread_t app_pthd_id;

static bool TThread_app_ChkIsThread() {
	return app_pthd_id != pthread_self();
}

static void TThread_PTH_SYNC_SIG(int sig) {
//   printf("%s(%d): Tpthd_PTH_SYNC_SIG pid=%d thd=0x%lx\n", FL, getpid(), pthread_self());
}

void REGP1 TThread_lock(TThread *self) {
	if (self) pthread_mutex_lock(&self->thd_pv->mutex);
}

void REGP1 TThread_unlock(TThread *self) {
	if (self) pthread_mutex_unlock(&self->thd_pv->mutex);
}
#endif
/*_____________________________ Sync ____________________________________*/
void TThread_Sync_app() {
// printf("%s(%d): %3d Sync_app %d\n", FL, GetTime_ms() - appDat.app_start_ms, appDat.Sleep);
	atomic_or(&appDat.appFu, appFu_select_t0_single);
#ifdef _OS_WIN32
	if (appDat.Sleep) SetEvent(appDat.app_Event);
#else
	if (appDat.Sleep) kill(appDat.self_pid, SIGCHLD);
#endif
}

void TThread_Sync_pth(TThread *self) {
  TThread_pv *pv;

	pv = self->thd_pv;
// printf("%s(%d): %3d Sync_pth %d %d %d 0x%08x\n", FL, GetTime_ms() - appDat.app_start_ms, pv->Sleep, pv->Start, pv->Stop, (int)pv->id);
	if (pv->Start && !pv->Stop && pv->id && pv->Sleep) {
#ifdef _OS_WIN32
	    SetEvent(pv->Event);
#else
	    pthread_kill(pv->id, SIGCHLD);
#endif
	}
}
/*_____________________________ MSG _____________________________________*/
void res_thdMSG(TThread_MSG *msg) {
	if (!msg) return;
	if (msg->u.type & uthd_t_MemFree) {
	    switch (msg->u.type & uthd_t_Mask) {
	    case uthd_t_pc:
	    case uthd_t_p:
		xfree(msg->u.p);
		break;
	    case uthd_t_ttc:
		Tfree(msg->u.ttc);
		break;
	    }
	}
	Fill0(msg, sizeof(TThread_MSG));
}

static void free_thdMsgBuf(TthdMsgBuf *buf) {

	while (buf->OVthdMSG) {
	    TOVthdMSG *ov = buf->OVthdMSG;
	    buf->OVthdMSG = ov->next;
	    res_thdMSG(&ov->thdMSG);
	    xfree(ov);
	}
	while (buf->msg_in != buf->msg_out) {
	    res_thdMSG(&buf->thdMSG[buf->msg_in]);
	    buf->msg_in++;
	    if (buf->msg_in >= MaxMsgBuf) buf->msg_in = 0;
	}
	buf->msg_in = buf->msg_out = 0;
}

static bool get_thd_pMSG(TthdMsgBuf *buf, TThread_MSG *msg) {
	if (buf->msg_in == buf->msg_out) return false;
	pnmove(msg, &buf->thdMSG[buf->msg_out], sizeof(TThread_MSG));
	buf->msg_out++;
	if (buf->msg_out >= MaxMsgBuf) buf->msg_out = 0;

	TOVthdMSG *ov = buf->OVthdMSG;
	if (ov) {
	    pnmove(&buf->thdMSG[buf->msg_in], &ov->thdMSG, sizeof(TThread_MSG));
	    buf->OVthdMSG = ov->next;
	    xfree(ov);
	    buf->msg_in++;
	    if (buf->msg_in >= MaxMsgBuf) buf->msg_in = 0;
	}
	return true;
}

static void add_thd_pMSG(TthdMsgBuf *buf, TThread_MSG *msg) {
  TThread_MSG *pmsg;

	int idx = buf->msg_in;
	idx++;
	if (idx >= MaxMsgBuf) idx = 0;
	if (idx == buf->msg_out) {
	    TOVthdMSG *ov, **ppov;
	    ov = xcalloc(sizeof(TOVthdMSG));
	    ppov = &buf->OVthdMSG;
	    while (*ppov) ppov = &(*ppov)->next;
	    *ppov = ov;
	    pmsg = &ov->thdMSG;
	} else {
	    pmsg = &buf->thdMSG[buf->msg_in];
	    buf->msg_in++;
	    if (buf->msg_in >= MaxMsgBuf) buf->msg_in = 0;
	}
	pnmove(pmsg, msg, sizeof(TThread_MSG));
}

void app2thd_pMSG(TThread *self, int send_sync, TThread_MSG *msg) {
  TThread_pv *pv;

	TThread_lock(self);
	pv = self->thd_pv;
	if (pv->Stop) {
	    send_sync = false;
	} else if (msg->thdMSG != thdMSG_OnlyEvent) {
	    add_thd_pMSG(&pv->thdMsgBuf, msg);
	}
	TThread_unlock(self);
	if (send_sync) TThread_Sync_pth(self);
}

void app2thd_MSG(TThread *self, int send_sync, int thdMSG, int tag) {
  TThread_MSG msg;

	Fill0(&msg, sizeof(TThread_MSG));
	msg.thdMSG = thdMSG;
	msg.tag = tag;
	app2thd_pMSG(self, send_sync, &msg);
}

/*__________________ TThread to App -> App Read __________________________*/
static void thd2app_cls_EV(TThread *self) {
  TThread **ppthd;

	ppthd = &app_private_thd;
	while (*ppthd) {
	    if (*ppthd == self) {
		*ppthd = self->app_private_thd;
		self->app_private_thd = NULL;
		free_thdMsgBuf(&self->thd_pv->appMsgBuf);
		return;
	    }
	    ppthd = &(*ppthd)->app_private_thd;
	}
}

static void thd2app_add_EV(TThread *self) {
  TThread **ppthd;

	ppthd = &app_private_thd;
	while (*ppthd) {
	    if (*ppthd == self) return;
	    ppthd = &(*ppthd)->app_private_thd;
	}
	*ppthd = self;
}

/* in AppHandleMessageRun() 
   if (appDat.thd_private_EV) appDat.thd_private_EV();
*/
static void thd2app_private_EV() {
  TThread *thd;
  TthdMsgBuf *buf;
  TThread_MSG msg;
  TThread_pv *pv;


	if (thd_private_EV_lock) {
	    AppSetWait1_single();
	    appDat.thd_private_EV = NULL;
	    return;
	}
	thd_private_EV_lock = true;

	appMutex_lock();
	thd = app_private_thd;
	while (thd) {
	    thd->thd_pv->flag |= THDflag_EVsend;
	    thd = thd->app_private_thd;
	}
loop:
	thd = app_private_thd;
	while (thd) {
	    pv = thd->thd_pv;
	    if (pv->IsWaitThdToAppMSG) {
		pv->IsWaitThdToAppMSG = false;
		if (thd->OnMsgEvent) {
		    appDat.thd_private_EV = thd2app_private_EV;
		    appMutex_unlock();
		    pv->ResWaitThdToAppMSG = thd->OnMsgEvent(thd, 
						&pv->WaitThdToAppMSG, 
						&pv->WaitThdToAppMSG_Result);
		    pv->IsWaitThdToAppMSG_Result = true;
		    TThread_Sync_pth(thd);
		    appMutex_lock();
		    goto loop;

		} else {
		    pv->ResWaitThdToAppMSG = 0;
		    pv->IsWaitThdToAppMSG_Result = true;
		    goto loop;
		}
	    }

	    buf = &pv->appMsgBuf;
	    if (pv->flag & THDflag_EVsend) {
		pv->flag &= ~THDflag_EVsend;

		if (get_thd_pMSG(buf, &msg)) {
		    if (thd->OnMsgEvent) {
			appDat.thd_private_EV = thd2app_private_EV;
			appMutex_unlock();
			thd->OnMsgEvent(thd, &msg, NULL);
/*
 int i = buf->msg_in - buf->msg_out;
 if (i < 0) i = -i;
 if (i > 0) printf("%s(%d): chk_Wait_msg msg stk = %d  ---------- \n", FL, i);
*/
			appMutex_lock();
			goto loop;
		    }
		}
	    }

	    if (buf->msg_in == buf->msg_out) thd2app_cls_EV(thd);
	    thd = thd->app_private_thd;
	}

	if (!app_private_thd) {
	    appDat.thd_private_EV = NULL;
	} else {
	    appDat.thd_private_EV = thd2app_private_EV;
	}
	appMutex_unlock();
	thd_private_EV_lock = false;
}
/*_____________________________ MSG to app ______________________________*/
void thd2app_pMSG(TThread *self, int send_sync, TThread_MSG *msg) {
  TThread_pv *pv;

//	if (appDat.AppExit) return;
	if (appDat.AppExit || !self->OnMsgEvent || self->thd_pv->free) return;
	appMutex_lock();
	pv = self->thd_pv;
	if (msg->thdMSG != thdMSG_OnlyEvent) {
	    add_thd_pMSG(&pv->appMsgBuf, msg);
	    thd2app_add_EV(self);
	    appDat.thd_private_EV = thd2app_private_EV;
	}
	appMutex_unlock();
	if (send_sync) TThread_Sync_app();
}

void thd2app_MSG(TThread *self, int send_sync, int thdMSG, int tag) {
  TThread_MSG msg;

	Fill0(&msg, sizeof(TThread_MSG));
	msg.thdMSG = thdMSG;
	msg.tag = tag;
	thd2app_pMSG(self, send_sync, &msg);
}

int thd2app_MSG_Wait(TThread *self, int thdMSG, int tag, TThread_MSG *ret_msg) {
  TThread_pv *pv;

	if (appDat.AppExit || !self->OnMsgEvent || self->thd_pv->free) return 0;
	pv = self->thd_pv;
	Fill0(&pv->WaitThdToAppMSG, sizeof(TThread_MSG));
	Fill0(&pv->WaitThdToAppMSG_Result, sizeof(TThread_MSG));
	pv->WaitThdToAppMSG.thdMSG = thdMSG;
	pv->WaitThdToAppMSG.tag = tag;

	appMutex_lock();
	thd2app_add_EV(self);
	appDat.thd_private_EV = thd2app_private_EV;
	pv->ResWaitThdToAppMSG = 0;
	pv->IsWaitThdToAppMSG_Result = false;
	pv->IsWaitThdToAppMSG = true;
	appMutex_unlock();

	TThread_Sync_app();

	while (!appDat.AppExit && !pv->free) {
	    if (pv->IsWaitThdToAppMSG_Result) {
		if (ret_msg) {
		    pnmove(ret_msg, &pv->WaitThdToAppMSG_Result, sizeof(TThread_MSG));
		} else {
		    res_thdMSG(&pv->WaitThdToAppMSG_Result);
		}
		return pv->ResWaitThdToAppMSG;
	    }
	    pv->Sleep = true;
#ifdef _OS_WIN32
  HANDLE EventHandles[2];
	    EventHandles[0] = pv->Event;
	    EventHandles[1] = 0;
	    WaitForMultipleObjects(1, EventHandles, false, 10);
#else
	    usleep(10000);
#endif
	    pv->Sleep = false;
	}
//	longjmp (self->thd_pv->ptAbort, 0);
	return 0;
}
/*_____________________________ Wait ____________________________________*/
static bool chk_Wait_msg(TThread *self) {
  TthdMsgBuf *buf;
  bool result = false;

	buf = &self->thd_pv->thdMsgBuf;
	if (buf->msg_in != buf->msg_out) {
	    TThread_lock(self);
	    result = get_thd_pMSG(buf, &self->msg);
/*
 int i = buf->msg_in - buf->msg_out;
 if (i < 0) i = -i;
 if (i > 1) printf("chk_Wait_msg msg stk = %d\n", i);
*/
	    TThread_unlock(self);
 	}
	return result;
}

int TThread_Wait(TThread *self, dword TimeOut_us) {
  TThread_pv *pv;
#ifdef _OS_WIN32
  HANDLE EventHandles[2];
	TimeOut_us /= 1000;
#else
  struct timeval totv;
#endif

	res_thdMSG(&self->msg);
	pv = self->thd_pv;
	if (chk_Wait_msg(self)) return thdWait_msg;
	pv->Sleep = TimeOut_us > 0;

#ifdef _OS_WIN32
	EventHandles[0] = pv->Event;
	EventHandles[1] = self->fd;
	int hCount = 1;
	if (self->fd) hCount++;
	DWORD Signaled = WaitForMultipleObjects(hCount, EventHandles, false, TimeOut_us);
	pv->Sleep = false;
	if (Signaled == (WAIT_OBJECT_0 + 1)) {
	    return thdWait_fd;
	}
#else
	totv.tv_sec = 0;
	totv.tv_usec = TimeOut_us;
	if (self->fd > 0) {
	    FD_ZERO (&pv->fdset_in);
	    FD_SET(self->fd, &pv->fdset_in);
	    if (select(FD_SETSIZE, &pv->fdset_in, NULL, 0, &totv) > 0) {
		pv->Sleep = false;
		if (FD_ISSET(self->fd, &pv->fdset_in)) {
		    return thdWait_fd;
		}
	    }
	    pv->Sleep = false;
	} else {
	    if (TimeOut_us > 0) select(0, NULL, NULL, 0, &totv);
	    pv->Sleep = false;
	}
#endif
	if (chk_Wait_msg(self)) return thdWait_msg;
	return thdWait_TimeOut;
}
/*_____________________________ Execute _________________________________*/
#ifdef _OS_WIN32
static DWORD WINAPI TThread_Execute(TThread *self) {
#else
static void *TThread_Execute(TThread *self) {
#endif
	if (setjmp(self->thd_pv->ptAbort)) goto ende;

	if (self->DoExecute) self->DoExecute(self);
// printf("%s(%d): TThread_Execute Ende\n", FL); fflush(stdout);
ende:
// printf("%s(%d): TThread_Execute Ende ptAbort thd=0x%08lx pthd=0x%08x\n", FL, self->thd_pv->id, (int)self); fflush(stdout);
	TThread_lock(self);
	free_thdMsgBuf(&self->thd_pv->thdMsgBuf);
	self->thd_pv->id = 0;
	self->thd_pv->Stop = true;
	TThread_unlock(self);
	if (self->thd_pv->free) Tfree(self);
	TThread_Sync_app();

#ifdef _OS_WIN32
	return 0;
#else
	return NULL;
#endif
}

/*_____________________________ Stop ____________________________________*/
bool TThread_Start(TThread *self) {
  TThread_pv *pv;

	if (!ttc_chk_is_wtyp(self, wtyp_Thread)) return false;
	pv = self->thd_pv;
	if (pv->Start && pv->Stop && !pv->id) {
	    pv->Start = 0;
	    pv->Stop = 0;
	}

	if (!pv->Start) {
#ifdef _OS_WIN32
	    pv->handle = CreateThread(NULL, 0, (void*)TThread_Execute, self, 0, &pv->id);
	    if (!pv->handle) {
		pv->id = 0;
		appError(EMSG_ERR1|EMSG_SYS, "%s(%d): pthread_create", FL);
		return false;
	    }
#else
	    if (pthread_create(&pv->id, NULL, (void*)TThread_Execute, self)) {
		pv->id = 0;
		appError(EMSG_ERR1|EMSG_SYS, "%s(%d): pthread_create", FL);
		return false;
	    }
	    pthread_detach(pv->id);
#endif
	    pv->Start = true;
	}
	return true;
}

/*_____________________________ Stop ____________________________________*/
void TThread_Stop(TThread *self) {
  TThread_pv *pv;

	if (!ttc_chk_is_wtyp(self, wtyp_Thread)) return;
	pv = self->thd_pv;
	if (pv->Start && !pv->Stop && pv->id) {
	    app2thd_MSG(self, true, thdMSG_CloseProcess, 0);
	    pv->Stop = true;
	}
}

bool TThread_ChkRun(TThread *self) {
  TThread_pv *pv;

	if (!ttc_chk_is_wtyp(self, wtyp_Thread)) return false;
	pv = self->thd_pv;
	if (pv->Start && pv->id) return true;
	return false;
}

bool TThread_StopAndWait(TThread *self, int ms, int fkt) {
  dword tms;

	if (!ttc_chk_is_wtyp(self, wtyp_Thread)) return false;
	TThread_Stop(self);
	if (ms > 0) {
	    if (ms > 10000) ms = 10000;
	    tms = GetTime_ms() + ms;

	    while (!cpTime(tms, GetTime_ms()) && TThread_ChkRun(self)) {
#ifdef _OS_WIN32
		AppHandleMessage();
#else
		if (fkt == thdSAW_FKT_USE_APP_HMSG) {
		    AppHandleMessage();
		} else {
		    usleep(1000);
		}
#endif
	    }
	}
	return TThread_ChkRun(self);
}

/*_____________________________ Destroy _________________________________*/
static bool TThread__Destroy(TThread *self) {
  TThread_pv *pv;

/*
	pv = self->thd_pv;
	if (!TThread_app_ChkIsThread()) {
	    if (pv->Start && !pv->Stop) {
		pv->free = true;
		thd2app_cls_EV(self);
		TThread_Stop(self);
	    } else {
		fprintf(stderr, "%s(%d) ERROR: TThread 2x Free Self ???\n", FL);
	    }
	    return true;
	}
*/

	pv = self->thd_pv;
	if (pv->Start && !pv->Stop) {
//   printf("%s(%d): TThread__Destroy Send STOP\n", FL);
	    if (TThread_app_ChkIsThread())  {
		fprintf(stderr, "%s(%d) ERROR: TThread Free Self ???\n", FL);
	    }

	    pv->free = true;
	    thd2app_cls_EV(self);
	    TThread_Stop(self);
	    return true;
	}

//   printf("%s(%d): TThread__Destroy %d id=0x%lx thd=0x%lx\n", FL, pv->free, pv->id, pthread_self());

	if (pv->id) {
	    TThread_lock(self);
	    if (pv->id) {
		if (TThread_app_ChkIsThread())  {
		    fprintf(stderr, "%s(%d) ERROR: TThread Free Self ???\n", FL);
		}
		pv->free = true;
		appMutex_lock();
		TThread_unlock(self);
		thd2app_cls_EV(self);
		appMutex_unlock();
		return true;
	    }
	    TThread_unlock(self);
	}
/*
	if (!TThread_app_ChkIsThread())  {
	    fprintf(stderr, "%s(%d) ERROR: TThread Free is not Self\n", FL);
	}
*/
	thd2app_cls_EV(self);
	free_thdMsgBuf(&self->thd_pv->thdMsgBuf);
	free_thdMsgBuf(&self->thd_pv->appMsgBuf);

#ifdef _OS_WIN32
	DeleteCriticalSection(&self->thd_pv->mutex);
	CloseHandle(self->thd_pv->Event);
#else
	pthread_mutex_destroy(&self->thd_pv->mutex);
#endif
	ppfree(&self->thd_pv);
	return false;
}

/*_____________________________ Create __________________________________*/
void *TThread_Create_init(int tcsize, int wtyp) {
  TThread *self;

	if (!wtyp) wtyp = wtyp_Thread;
	appMutex_init();
	self = ttc_create(tcsize, wtyp);
	self->destroy = (TDESTROY)TThread__Destroy;
	self->thd_pv = xcalloc(sizeof(TThread_pv));

#ifdef _OS_WIN32
	if (!appDat.app_Event) {
	    appDat.app_Event = CreateEvent(NULL, false, false, NULL);
	    AddAppMsgEvent(appDat.app_Event);
	}
	self->thd_pv->Event = CreateEvent(NULL, false, false, NULL);
	InitializeCriticalSection(&self->thd_pv->mutex);
#else
	if (!app_pthd_id) {
	    app_pthd_id = pthread_self();
	    app_ChkIsThread = TThread_app_ChkIsThread;
	    struct sigaction action, old_action;
	    sigfillset (&action.sa_mask);
	    action.sa_handler = TThread_PTH_SYNC_SIG;
	    action.sa_flags = 0;
	    sigaction (SIGCHLD, &action, &old_action);
	}
	pthread_mutex_init(&self->thd_pv->mutex, NULL);
#endif
//   printf("%s(%d): TThread_Create pid=%d thd=0x%lx\n", FL, getpid(), pthread_self());
	return self;
}

TThread *TThread_Create() {
	return TThread_Create_init(sizeof(TThread), 0);
}
