/*
 Copyright (C) 2010-2022 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"
#include "sysbaseio.h"
#include <unistd.h>
#include <sys/stat.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>

static bool chk_dev_file(const char *DevFile) {
	if (!DevFile || !pcncmp(DevFile, "/dev/", 5)) {
	    errno = 1;
	    return false;
	}
	return true;
}

static char *get_lock_file_name(const char *devname) {
  char *result, *pc;

	if (!*devname) return NULL;
	result = pcadd("/var/lock/LCK..", devname);
	if (pclen(result) >= 1024) {
	    xfree(result);
	    return NULL;
	}
	pc = result + 12;
	while (*pc) {
	    if (*pc == '/') *pc = ':';
	    pc++;
	}
	return result;
}

static pid_t get_lock_pid(const char *lock_file) {
  FileHandle fd;
  char buf[64];
  struct stat64 tmpstat;

	if (stat64(lock_file, &tmpstat)) return 0;
	fd = open64(lock_file, O_RDONLY);
	if (fd < 1) return -1;
	buf[0] = 0;
	read(fd, buf, 64);
	close(fd);
	return pctoint(buf);
}

static bool get_pid_run(pid_t pid) {
	if (!kill(pid, 0)) return true;
	return errno != ESRCH;
}

pid_t DevTestLock(const char *DevFile) {
  pid_t result;
  char *lock_file;
  struct stat64 tmpstat;

	if (!chk_dev_file(DevFile)) return -1;
	if (stat64(DevFile, &tmpstat)) return -errno;
	if (euidaccess(DevFile, W_OK) == -1 ) return -errno;

	lock_file = get_lock_file_name(DevFile + 5);
	result = get_lock_pid(lock_file);
// printf("%s(%d): DevTestLock lock_file=%s pid=%d self=%d\n", FL, lock_file, result, getpid());
	if ((result > 0) && (result == getpid())) result = 0;

	if ((result > 0) && !get_pid_run(result)) {
	    unlink(lock_file);
	    result = 0;
	}
	xfree(lock_file);
	return result;
}

pid_t DevLock(const char *DevFile) {
  pid_t result;
  size_t lg;
  FileHandle fd;
  char *pc, *lock_file;
  struct stat64 tmpstat;
  char buf[20];

	if (!chk_dev_file(DevFile)) return -1;
	if (stat64(DevFile, &tmpstat)) return -errno;
	if (euidaccess(DevFile, W_OK) == -1 ) return -errno;

	lock_file = get_lock_file_name(DevFile + 5);
	result = get_lock_pid(lock_file);
// printf("%s(%d): DevLock   lock_file=%s pid=%d self=%d\n", FL, lock_file, result, getpid());
	if ((result > 0) && (result == getpid())) result = 0;
	if ((result > 0) && !get_pid_run(result)) result = 0;

	if (!result) {
	    fd = open64(lock_file, O_CREAT | O_TRUNC | O_WRONLY, 0644);
	    if (ChkFileHandleOK(fd)) {
		pc = buf;
		*pc = ' '; pc++;
		inttoppc(&pc, getpid());
		*pc = '\n'; pc++; *pc = 0;
		lg = write(fd, buf, pclen(buf));
		if (lg > 0) result = 0;
		close(fd);
	    }
	}
	xfree(lock_file);
	return result;
}

pid_t DevUnlock(const char *DevFile, pid_t pid) {
  pid_t result;
  char *lock_file;

	if (!chk_dev_file(DevFile)) return -1;
	lock_file = get_lock_file_name(DevFile + 5);
	if (!pid) pid = getpid();
	result = get_lock_pid(lock_file);
// printf("%s(%d): DevUnlock lock_file=%s pid=%d self=%d\n", FL, lock_file, result, getpid());
	if ((result > 0) && (result == pid)) result = 0;
	if (!result) {
	    unlink(lock_file);
	}
	xfree(lock_file);
	return result;
}
