/*
 Copyright (C) 2010-2022 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#define IS_CLASS_LIB

#include "appbase.h"
#include "trs232ald.h"

/*_____________________________ Misc _________________________________*/
static char Errbuf[100];

const char *Trs232ald_GetErrorStr(dword orig_bef, dword bef, const byte *data, 
				  size_t lg, bool unknown) {
  const char *msg;
  char *pc;

	if (bef == RSBEF_ERROR) {
	    msg = NULL;
	    switch (*data) {
	    case RSERR_CRC:		msg = "ERR_CRC"; break;
	    case RSERR_BUFLEN:		msg = "ERR_BUFLEN"; break;
	    case RSERR_UNKNOWN_BEF:	msg = "ERR_UNKNOWN_BEF"; break;
	    case RSERR_BEF_SIZE:	msg = "ERR_BEF_SIZE"; break;
	    case RSERR_ADDR_FAULT:	msg = "ERR_ADDR_FAULT"; break;
	    case RSERR_LEN_OVERFLOW:	msg = "ERR_LEN_OVERFLOW"; break;
	    case RSERR_READ_ONLY:	msg = "ERR_READ_ONLY"; break;
	    }

	    pc = Errbuf;
	    if (msg) {
		if (!orig_bef) return msg;
		ppcmove(&pc, msg);
		goto chk_orig_bef;

	    } else if (unknown) {
		ppcmove(&pc, "UNKNOWN ERROR  err=");
		hexto0xppc(&pc, data[0], 2);
		if (lg > 1) {
		    *pc = '/'; pc++;
		    hexto0xppc(&pc, data[1], 2);
		}
		goto chk_orig_bef;
	    }
	} else if (unknown) {
	    pc = Errbuf;
	    ppcmove(&pc, "bef=");
	    hexto0xppc(&pc, bef, 2);
chk_orig_bef:
	    if (orig_bef) {
		ppcmove(&pc, "  orig.bef=");
		hexto0xppc(&pc, orig_bef, 2);
	    }
	    return Errbuf;
	}
	return NULL;
}

/* ___________________________ TimeOut __________________________________ */
static void OnTimeOutTimer(Trs232ald *self) {
	if (self->TxDRepMaxCount > 0) {
	    if (self->TxDRepCount >= self->TxDRepMaxCount) {
/*
		if (!self->OnERROR || 
		    !self->OnERROR(self, rs232ERR_TIMEOUT, NULL)) {
		    appError(EMSG_ERR1, "rs232 RxD TimeOut");
*/
		if (!Trs232_SErr((Trs232*)self, rs232ERR_TIMEOUT, EMSG_ERR1, "rs232 RxD TimeOut")) {
		    exit(EXIT_FAILURE);
		}
	    } else {
		self->TxDRepCount++;
/*
		if (!self->OnERROR || 
		    !self->OnERROR(self, rs232ERR_TIMEOUT_REP, NULL)) {
		    appError(EMSG_WARN, "rs232 RxD TimeOut count=%d", self->TxDRepCount);
		}
*/
		Trs232_SErr((Trs232*)self, rs232ERR_TIMEOUT_REP, EMSG_WARN, "rs232 RxD TimeOut count=%d", self->TxDRepCount);
		Trs232_TxDBuf_Rep((Trs232*)self);
	    }
	}
}

static dword GetCtlCRC(unsigned int dlen, unsigned int bef, unsigned int ctl) {
	return ((bef + dlen) ^ ~((bef + dlen + (ctl & 0xf0)) >> 4)) & 0x0f;
}

static dword GetCRC(const byte *data, unsigned int dlen) {
  int i;
  unsigned int result;
  const byte *pb;

//sum = 0x39 + Bef + len + ctl + data[n]
	pb = data + 1;
	result = 0x39;
	for (i=dlen+3;i>0;i--) {
	    result += *pb;
	    pb++;
	}
	return result & 0xff;
}

static bool OnRxD(Trs232ald *self, byte *data, size_t lg) {
  unsigned int dlen, bef, ctl, sum;
  byte *pdata;

#if 0
  printf("%s(%d): lg=%3d ", FL, lg);
  int mx = lg;
  if (mx > 18) mx = 18;
  for (int i=0;i<mx;i++) printf(" %02x", data[i]);
  printf("\n");
#endif
	if (data[0] != 0xff) goto clsbuf;

	if (lg < 2) return false;
	bef  = data[1];
	if ((bef == 0) || (bef == 0xff)) goto clsbuf;

	if (lg < 3) return false;
	dlen  = data[2];
	if (dlen > 252) goto clsbuf;

	if (lg < 4) return false;
	ctl  = data[3];
	if ((ctl & 0x0f) != GetCtlCRC(bef, dlen, ctl)) {
// printf("%s(%d): ctl CRC RxD=0x%02x <> PC=0x%02x  0x%02x 0x%02x 0x%02x\n", FL, (ctl & 0x0f), GetCtlCRC(bef, dlen, ctl), bef, dlen, ctl);
clsbuf:
	    self->RxDidx = 0;
	    return false;
	}
	if (ctl & CTL_4x) dlen <<= 2; //4x

	if (lg < (dlen + 5)) return false;

	sum = GetCRC(data, dlen);
	if (sum != data[dlen + 4]) {
// printf("%s(%d): sum=0x%02x <> 0x%02x\n", FL, sum, data[dlen + 4]);
	    self->RxDidx = 0;
	    if (self->TxDRepMaxCount > 0) {
		if (self->TxDRepCount >= self->TxDRepMaxCount) {
/*
		    if (!self->OnERROR || 
			!self->OnERROR(self, rs232ERR_RxD_CRC, NULL)) {
			appError(EMSG_ERR1, "rs232 RxD CRC");
			exit(EXIT_FAILURE);
		    }
*/
		    if (!Trs232_SErr((Trs232*)self, rs232ERR_RxD_CRC, EMSG_ERR1, "rs232 RxD CRC")) {
			exit(EXIT_FAILURE);
		    }
		} else {
		    self->TxDRepCount++;
/*
		    if (!self->OnERROR || 
			!self->OnERROR(self, rs232ERR_RxD_CRC_REP, NULL)) {
			appError(EMSG_WARN, "rs232 RxD CRC count=%d  sum=0x%02x <> 0x%02x", self->TxDRepCount, sum, data[dlen + 4]);
		    }
*/
		    Trs232_SErr((Trs232*)self, rs232ERR_RxD_CRC_REP, EMSG_WARN, "rs232 RxD CRC count=%d  sum=0x%02x <> 0x%02x", self->TxDRepCount, sum, data[dlen + 4]);
		    Trs232_TxDBuf_Rep((Trs232*)self);
		}
	    }
	    return false;
	}

	pdata = data + 4;
	if ((bef == RSBEF_ERROR) && (dlen == 1) && 
	    (*pdata == RSERR_CRC)) //Err TxD Summe
	{
	    if (self->TxDRepMaxCount > 0) {
		if (self->TxDRepCount >= self->TxDRepMaxCount) {
/*
		    if (!self->OnERROR || 
			!self->OnERROR(self, rs232ERR_TxD_CRC, NULL)) {
			appError(EMSG_ERR1, "Slave has CRC error", self->TxDRepCount);
			exit(EXIT_FAILURE);
		    }
*/
		    if (!Trs232_SErr((Trs232*)self, rs232ERR_TxD_CRC, EMSG_ERR1, "Slave has CRC error")) {
			exit(EXIT_FAILURE);
		    }
		} else {
		    self->TxDRepCount++;
/*
		    if (!self->OnERROR || 
			!self->OnERROR(self, rs232ERR_TxD_CRC_REP, NULL)) {
			appError(EMSG_WARN, "Slave has CRC error count=%d", self->TxDRepCount);
		    }
*/
		    Trs232_SErr((Trs232*)self, rs232ERR_RxD_CRC_REP, EMSG_WARN, "Slave has CRC error count=%d", self->TxDRepCount);
		    Trs232_TxDBuf_Rep((Trs232*)self);
		}
	    }
	} else {
	    if (self->OnRxDData) {
		self->OnRxDData(self, bef, pdata, dlen);
	    }
	}
	return true;
}

/* ---------------------------- TxD ---------------------------------- */

bool rs232ald_SendData(Trs232ald *self, unsigned int bef, const byte *data, size_t lg) {
  unsigned int sbef, ctl;
  byte buf[520];

	if (!self) return false;
	Trs232_ClrTimeOut((Trs232*)self);
	if (!ChkFileHandleAkt(self->fd)) return false;

	if (!data) lg = 0;
	if (lg > 512)  return false;

	sbef = bef & 0xff;
	ctl = 0;
	if (bef & RSBEF_4xMODE) {
	    ctl = CTL_4x;
	}

	if (sbef == RSBEF_SYNC) self->TxD_tgl = 0;
	ctl |= self->TxD_tgl;
	buf[0] = 0xff;
	buf[1] = bef;
	buf[2] = lg;
	buf[3] = ctl | GetCtlCRC(bef, lg, ctl);
	pnmove(buf + 4, data, lg);
	buf[lg + 4] = GetCRC(buf, lg);

	if (!Trs232_TxD((Trs232*)self, buf, lg + 5)) return false;
	self->LastTxDBef = bef;
	self->TxD_tgl ^= CTL_TOGGLE;
	return true;
}

/*_____________________________ Destroy _________________________________*/
bool Trs232ald__Destroy(Trs232ald *self) {
	return Trs232__Destroy((Trs232*)self);
}
/*_____________________________ Create __________________________________*/
void *Trs232ald_Create_struct_init(int tcsize, int wtyp) {
  Trs232ald *self;

	self = Trs232_Create_struct_init(tcsize, wtyp);
	self->destroy = (TDESTROY)Trs232ald__Destroy;
	if (!self->TxDbuf) {
	    self->TxDsize = RS232MAXBUF;
	    self->TxDbuf = xmalloc(self->TxDsize);
	}
	self->OnRxD = OnRxD;
	self->TxDRepMaxCount = 4;
	self->OnTimeOutTimer = OnTimeOutTimer;
	Trs232_InitTimeOut((Trs232*)self, 400);
	return self;
}

Trs232ald *Trs232ald_Create() {
	return Trs232ald_Create_struct_init(sizeof(Trs232ald), wtyp_Trs232ald);
}
