/*
 Copyright (C) 2010-2024 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "image.h"
#include "media_io.h"
#include "video_ioc.h"

#define pmxWR_bppMAXnr		9

#define pmxWR_bppMask		0x0f0
#define pmxWR_bppDefault	0x000
#define pmxWR_bpp6		0x010
#define pmxWR_bpp7		0x020
#define pmxWR_bpp8		0x030
#define pmxWR_bpp9		0x040
#define pmxWR_bpp10		0x050
#define pmxWR_bpp12		0x060
#define pmxWR_bpp15		0x070
#define pmxWR_bpp16		0x080
#define pmxWR_bpp18		0x090

typedef struct {
  const byte kr, kg, kb;
} Tkrgb;

const Tkrgb Tkrgb_arr[pmxWR_bppMAXnr + 1] = {
  { 0,  0,  0}, //default
  { 5,  5,  5}, //pmxWR_bpp6
  { 5,  9,  5}, //pmxWR_bpp7
  { 9,  9,  5}, //pmxWR_bpp8
  { 9,  9,  9}, //pmxWR_bpp9
  { 9, 17,  9}, //pmxWR_bpp10
  {17, 17, 17}, //pmxWR_bpp12
  {33, 33, 33}, //pmxWR_bpp15
  {33, 65, 33}, //pmxWR_bpp16
  {65, 65, 65}  //pmxWR_bpp18
};

#ifdef MEM_CHK
static int LastMEM_CHK_Count;
static void AzMem() {
  int lMCt;  
  dword lMsize;
	MEM_CHK_Count(&lMCt, &lMsize);
	if (LastMEM_CHK_Count != lMCt) {
		LastMEM_CHK_Count = lMCt;
    printf("Debug %s: MEM_CHK_Count=%3d  size=%d\n", app_name, lMCt, lMsize);
	}
}
#else
 #define AzMem() 
#endif

static int read_extColMap(dword *quant_col_map, const char *val) {
  dword m, r, g, b;

	m = 0;
	while (m < 8) {
	    r = g = b = ppctoint(&val) & 0xff;
	    if ((*val == '.')||(*val == ',')) {
		val++;
		g = ppctoint(&val) & 0xff;
		if ((*val == '.')||(*val == ',')) {
		    val++;
		    b = ppctoint(&val) & 0xff;
		}
	    }
	    quant_col_map[m] = b|(g<<8)|(r<<16);
	    m++;
	    if (*val != ':') break; else val++;
	}
	return m;
}

static const char *Help_Msg = {
  "\n"
  "Usage: imgconv [options] file\n"
  "  input  *.xpm *.png"
#ifdef USE_JPEG
  " *.jpg"
#endif
  " *.bmp"
  " *.pnm *.ppm *.pam" 
  "\n"
  "  output *.xpm *.png *.jpg *.pnm *.pbm *.pgm *.ppm *.pam"
  "\n"
  "\n"
  "  -o outfile           output file\n"
  "  -b=nn                bpp 6,7,8,9,10,12,15,18\n"
  "  -c=nn                color count (quantize colors)\n"
  "  -m=r.g.b             add default map color (max.8) (quantize colors)\n"
  "                         -m=0.0.0:255.255.255 (black;white)\n"
  "                         or -m=0:255 (black;white)\n"
  "  -z=K                 zoom percent (max. 1..200%)\n"
  "  -Z=WWWxHHH           zoom pixel W*H (max. 1..200%)\n"
#ifdef USE_JPEG
  "  -q=nn                jpg output quality 1..100 (default 95)\n"
#endif
  "  -C=n                 Compression level 0..9 (only *.png)\n"
  "  --plain              *.pnm output txt plain\n"
  "  -h, --help           prints out this screen\n"
};

int main (int argc, char **argv) {
  TImage *img, *rsz_img;
  int i;
  char *val, *fn;
  char *InFile, *OutFile;
  int ColorCount, bpp, pic_bpp;
  int ZoomProz, ZoomW, ZoomH, jpg_quality, pnm_plain;
  dword default_quant_col_map[8];
  int sum, default_quant_count, Compress;
  dword WR_fu;
  TVideo_IOC *ioc_colorRGB;
  TVideo_Color_Korr *tvckr;
  bool help, err, verbose;

	InFile = NULL;
	OutFile = NULL;
	rsz_img = NULL;
	ZoomProz = 0;
	ZoomW = 0;
	ZoomH = 0;
	ColorCount = 0;
	Compress = 0;
	jpg_quality = 0;
	bpp = 0;
	pnm_plain = 0;
	default_quant_count = 0;
	verbose = true;
	help = false;
	err = false;
	i = 1;
	while (!err && !help && (i<argc)) {
	    val = argv[i];
	    if (*val == '-') {
		val++;
		switch (*val) {
		case 'b':
		    err = cmd_get_int(argc, argv, val, &i, &bpp);
		    break;
		case 'c':
		    err = cmd_get_int(argc, argv, val, &i, &ColorCount);
		    break;
		case 'C':
		    err = cmd_get_int(argc, argv, val, &i, &Compress);
		    if (!err) {
			if (Compress < 0) Compress = 0;
			if (Compress > 9) Compress = 9;
			Compress++;
		    }
		    break;
		case 'm':
		    val++;
		    if (*val == '=') val++;
		    default_quant_count = read_extColMap(default_quant_col_map, val);
		    break;
		case 'o':
		    val++;
		    if (*val == '=') {
			val++;
			OutFile = pcdup(val);
		    } else if (((i+1)<argc) && (*argv[i+1] != '-')) {
			i++;
			OutFile = pcdup(argv[i]);
		    } else err = true;
		    break;
		case 'q':
		    err = cmd_get_int(argc, argv, val, &i, &jpg_quality);
		    break;
		case 'z':
		    err = cmd_get_int(argc, argv, val, &i, &ZoomProz);
		    if (ZoomProz > 200) err = true;
		    break;
		case 'Z':
		    val++;
		    if (*val == '=') val++;
		    ZoomW = ppctoint((void*)&val);
		    if (*val == 'x') {
			val++;
			ZoomH = pctoint(val);
		    } else err = true;
		    break;
		case 'h':
		    help = true;
		    break;
		case '-':
		    val++;
// printf("%s(%d): %s\n", FL, val);
		    if (pcncmp(val, "help", 4)) {
			help = true;

		    } else if (pcncmp(val, "plain", 5)) {
			pnm_plain = fuWR_txtplain;

		    } else if (pcncmp(val, "quiet", 5)) {
			verbose = false;

		    } else err = true;
		    break;
		default: 
		    err = true;
		}
	    } else {
		InFile = val;
	    }
	    i++;
	}

	pic_bpp = pmxWR_bppDefault;
	if (bpp) {
	    switch (bpp) {
	    case 6:  pic_bpp = pmxWR_bpp6; break;
	    case 7:  pic_bpp = pmxWR_bpp7; break;
	    case 8:  pic_bpp = pmxWR_bpp8; break;
	    case 9:  pic_bpp = pmxWR_bpp9; break;
	    case 10: pic_bpp = pmxWR_bpp10; break;
	    case 12: pic_bpp = pmxWR_bpp12; break;
	    case 15: pic_bpp = pmxWR_bpp15; break;
	    case 16: pic_bpp = pmxWR_bpp16; break;
	    case 18: pic_bpp = pmxWR_bpp18; break;
	    default: err = true;
	    }
	}

	err = err || !InFile;
	if (help || err) {
	    printf(Help_Msg);
	    exit(err);
	}

	ThImageRD_Reg(6, &hImageRD_xpm, &hImageRD_png, &hImageRD_jpeg, 
			 &hImageRD_gif, &hImageRD_bmp, &hImageRD_pnm);

	ThImageWR_Reg(4, &hImageWR_xpm, &hImageWR_png, &hImageWR_jpeg,
			 &hImageWR_pnm);

// printf("%s(%d): %s c=%d ecm=%d -> %s\n", FL, InFile, ColorCount, default_quant_count, OutFile);

	fn = path_rel2abs(InFile, NULL);
	if (!TestIsFile(fn)) {
	    fprintf(stderr, "ERROR: file not fond  %s\n", InFile);
	    return 1;
	}

	img = TImage_Loader(fn, NULL, pxm_4b_24bpp);
	xfree(fn);
	if (!img) exit(1);

	if (verbose) {
	    sum = TImage_GetColorCount(img, pxgcc_RGB_ALL);
	    printf("IN:  w=%d h=%d colors=%d\n", img->width, img->height, sum);
	}
	if (!OutFile) {
	    return 0;
	}

	if (ZoomProz) {
	    ZoomW = img->width * ZoomProz / 100;
	    if (ZoomW < 2) ZoomW = 2;
	    ZoomH = img->height * ZoomProz / 100;
	    if (ZoomH < 2) ZoomH = 2;
	}
	if (ZoomW && ZoomH) {
	    if (ZoomW < 2) ZoomW = 2;
	    if (ZoomH < 2) ZoomH = 2;
	    rsz_img = TImage_dup_new_size(img, ZoomW, ZoomH);
	    if (rsz_img) {
		Tfree(img);
		img = rsz_img;
	    }
	}

	if (pic_bpp != pmxWR_bppDefault) {
  TMediaFrame fio;
  TIOCVideoF iof;

	    Fill0(&fio, sizeof(TFrame_IO));
	    Fill0(&iof, sizeof(TIOCF));
	    fio.org.v.iof = &iof;
	    fio.ret.v.iof = &iof;
	    iof.typ = DatType_TImage;
	    fio.org.v.t.img = img;
	    fio.ret.v.t.img = img;

	    ioc_colorRGB = TVideo_IOC_ColorRGB_Create();
	    tvckr = ioc_colorRGB->tvckr;

	    i = (pic_bpp & pmxWR_bppMask) >> 4;
	    if (i > pmxWR_bppMAXnr) i = 0;
	    tvckr->R_reduce = Tkrgb_arr[i].kr;
	    tvckr->G_reduce = Tkrgb_arr[i].kg;
	    tvckr->B_reduce = Tkrgb_arr[i].kb;

	    Video_IOC_SetColor(ioc_colorRGB, NULL);
	    Video_IOC_Get_Frame(ioc_colorRGB, &fio);
//	    img = fio.ret.v.t.img;
	}



  ThImageLoader *hLD;
  FILE *F;
	fn = path_rel2abs(OutFile, NULL);
	F = TImageIO_GetStream(fn, NULL, 0, &hLD, true);
	if (!F) exit(1);

	if (ColorCount > 0) {
	    TImage_GetColorCount(img, pxgcc_RGB_ALL);
	    if (img->cmap) {
		img->cmap->default_quant_count = default_quant_count;
		img->cmap->default_quant_col_map = (Tic32*)default_quant_col_map;
	    }
	    TImage_GetColorMap(img);
	    TImage_CutColorMap(img, ColorCount, 0);
	    TImage_ReMapColor(img);
	}

	if (verbose) {
	    sum = TImage_GetColorCount(img, pxgcc_RGB_ALL);
	    printf("OUT: w=%d h=%d colors=%d\n", img->width, img->height, sum);
	}

	WR_fu = 0;
	if (hLD->filetyp & filetyp_PNG) {
	    if (Compress) WR_fu |= (Compress << 8);
	    WR_fu |= fuWR_use_pallette;
	}
	if (hLD->filetyp & filetyp_JPEG) WR_fu |= jpg_quality << 8;
	if (hLD->filetyp & filetyp_PNM)  WR_fu |= pnm_plain;
	
	img->out_gp = pctogp(fn);
	img->out_fu = WR_fu;
	img->out_fp = F;
	hLD->ioproc(img);
	fclose(img->out_fp);
	gpFree(img->out_gp);

 xfree(OutFile);
 xfree(fn);

 Tfree(img);
 AzMem();

	return 0;
}
